"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.FieldsCollector = void 0;
const typescript_1 = __importDefault(require("typescript"));
const dynamodb_primitive_types_1 = require("./dynamodb_primitive_types");
const dynamodb_item_field_1 = require("./dynamodb_item_field");
const logger_1 = require("./logger");
class FieldsCollector {
    constructor(funcName, shouldLenientTypeCheck) {
        this.funcName = funcName;
        this.shouldLenientTypeCheck = shouldLenientTypeCheck;
    }
    collectFields(node, typeChecker, typeName, type) {
        if (node.typeArguments === undefined || node.typeArguments.length !== 1 || !node.typeArguments[0]) {
            throw new Error(`No type argument on ${this.funcName}(). Please put a type argument on the function`);
        }
        if (node.arguments.length < 1 || !node.arguments[0]) {
            throw new Error(`No argument on ${this.funcName}(). Please put an argument that has ${typeName} type on the function`);
        }
        const properties = typeChecker.getPropertiesOfType(type);
        return properties.map((prop) => {
            var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k;
            if (!prop.valueDeclaration ||
                !typescript_1.default.canHaveModifiers(prop.valueDeclaration) ||
                !this.isPropertyModifierInArgumentSuitableForDynamodbAttr(typescript_1.default.getModifiers(prop.valueDeclaration))) {
                return undefined;
            }
            const propName = prop.name;
            if (prop.valueDeclaration.kind !== typescript_1.default.SyntaxKind.Parameter &&
                prop.valueDeclaration.kind !== typescript_1.default.SyntaxKind.PropertySignature) {
                const msg = `a property "${propName}" of the type "${typeName}" doesn't have parameter kind; maybe the ${typeName} is not class of interface`;
                if (this.shouldLenientTypeCheck) {
                    (0, logger_1.warn)(msg);
                    return undefined;
                }
                throw new Error(msg);
            }
            const valueDeclSymbol = typeChecker.getTypeAtLocation(prop.valueDeclaration).symbol;
            const valueDeclSymbolName = valueDeclSymbol === null || valueDeclSymbol === void 0 ? void 0 : valueDeclSymbol.name;
            if (valueDeclSymbolName === 'Array') {
                const typeArgs = typeChecker.getTypeArguments(typeChecker.getTypeAtLocation(prop.valueDeclaration));
                const valueType = (0, dynamodb_primitive_types_1.dynamodbPrimitiveTypeFromTypeFlag)((_a = typeArgs[0]) === null || _a === void 0 ? void 0 : _a.flags);
                if (valueType === undefined) {
                    const valueTypeName = (_c = (_b = typeArgs[0]) === null || _b === void 0 ? void 0 : _b.symbol) === null || _c === void 0 ? void 0 : _c.name;
                    if (valueTypeName === 'Uint8Array') {
                        return new dynamodb_item_field_1.ArrayField(propName, new dynamodb_primitive_types_1.DynamodbPrimitiveType(dynamodb_primitive_types_1.DynamodbPrimitiveTypeKinds.Binary, dynamodb_primitive_types_1.DynamodbPrimitiveTypeFlags.NA));
                    }
                    if (valueTypeName === 'BigInt') {
                        return new dynamodb_item_field_1.ArrayField(propName, new dynamodb_primitive_types_1.DynamodbPrimitiveType(dynamodb_primitive_types_1.DynamodbPrimitiveTypeKinds.Number, dynamodb_primitive_types_1.DynamodbPrimitiveTypeFlags.BigInt));
                    }
                    const msg = `a property "${propName}" of the type "${typeName}" has unsupported type: "${valueTypeName}"`;
                    if (this.shouldLenientTypeCheck) {
                        (0, logger_1.warn)(msg);
                        return undefined;
                    }
                    throw new Error(msg);
                }
                return new dynamodb_item_field_1.ArrayField(propName, valueType);
            }
            if (valueDeclSymbolName == 'Set') {
                const typeArgs = typeChecker.getTypeArguments(typeChecker.getTypeAtLocation(prop.valueDeclaration));
                const valueType = (0, dynamodb_primitive_types_1.dynamodbPrimitiveTypeFromTypeFlag)((_d = typeArgs[0]) === null || _d === void 0 ? void 0 : _d.flags);
                if (valueType === undefined) {
                    const valueTypeName = (_f = (_e = typeArgs[0]) === null || _e === void 0 ? void 0 : _e.symbol) === null || _f === void 0 ? void 0 : _f.name;
                    if (valueTypeName === 'Uint8Array') {
                        return new dynamodb_item_field_1.SetField(propName, new dynamodb_primitive_types_1.DynamodbPrimitiveType(dynamodb_primitive_types_1.DynamodbPrimitiveTypeKinds.Binary, dynamodb_primitive_types_1.DynamodbPrimitiveTypeFlags.NA), this.shouldLenientTypeCheck);
                    }
                    if (valueTypeName === 'BigInt') {
                        return new dynamodb_item_field_1.SetField(propName, new dynamodb_primitive_types_1.DynamodbPrimitiveType(dynamodb_primitive_types_1.DynamodbPrimitiveTypeKinds.Number, dynamodb_primitive_types_1.DynamodbPrimitiveTypeFlags.BigInt), this.shouldLenientTypeCheck);
                    }
                    const msg = `a property "${propName}" of the type "${typeName}" has unsupported type: ${valueTypeName}`;
                    if (this.shouldLenientTypeCheck) {
                        (0, logger_1.warn)(msg);
                        return undefined;
                    }
                    throw new Error(msg);
                }
                return new dynamodb_item_field_1.SetField(propName, valueType, this.shouldLenientTypeCheck);
            }
            if (valueDeclSymbolName == 'Map') {
                const typeArgs = typeChecker.getTypeArguments(typeChecker.getTypeAtLocation(prop.valueDeclaration));
                const keyType = (0, dynamodb_primitive_types_1.dynamodbPrimitiveTypeFromTypeFlag)((_g = typeArgs[0]) === null || _g === void 0 ? void 0 : _g.flags);
                if (keyType === undefined) {
                    const msg = `a Map type property "${propName}" of the type "${typeName}" has non-string key type`;
                    if (this.shouldLenientTypeCheck) {
                        (0, logger_1.warn)(msg);
                        return undefined;
                    }
                    throw new Error(msg);
                }
                const valueType = (0, dynamodb_primitive_types_1.dynamodbPrimitiveTypeFromTypeFlag)((_h = typeArgs[1]) === null || _h === void 0 ? void 0 : _h.flags);
                if (valueType === undefined) {
                    const valueTypeName = (_k = (_j = typeArgs[1]) === null || _j === void 0 ? void 0 : _j.symbol) === null || _k === void 0 ? void 0 : _k.name;
                    if (valueTypeName === 'Uint8Array') {
                        return new dynamodb_item_field_1.MapField(propName, keyType, new dynamodb_primitive_types_1.DynamodbPrimitiveType(dynamodb_primitive_types_1.DynamodbPrimitiveTypeKinds.Binary, dynamodb_primitive_types_1.DynamodbPrimitiveTypeFlags.NA), this.shouldLenientTypeCheck);
                    }
                    if (valueTypeName === 'BigInt') {
                        return new dynamodb_item_field_1.MapField(propName, keyType, new dynamodb_primitive_types_1.DynamodbPrimitiveType(dynamodb_primitive_types_1.DynamodbPrimitiveTypeKinds.Number, dynamodb_primitive_types_1.DynamodbPrimitiveTypeFlags.BigInt), this.shouldLenientTypeCheck);
                    }
                    const msg = `a property "${propName}" of the type "${typeName}" has unsupported type: "${valueTypeName}"`;
                    if (this.shouldLenientTypeCheck) {
                        (0, logger_1.warn)(msg);
                        return undefined;
                    }
                    throw new Error(msg);
                }
                return new dynamodb_item_field_1.MapField(propName, keyType, valueType, this.shouldLenientTypeCheck);
            }
            if (((valueDeclSymbol === null || valueDeclSymbol === void 0 ? void 0 : valueDeclSymbol.flags) & typescript_1.default.SymbolFlags.TypeLiteral) === typescript_1.default.SymbolFlags.TypeLiteral) {
                const kvType = this.extractKeyValueTypesFromKeyValuePairMapSyntax(prop.valueDeclaration.getChildren());
                const keyTypeName = kvType === null || kvType === void 0 ? void 0 : kvType[0];
                const keyType = (0, dynamodb_primitive_types_1.dynamodbPrimitiveTypeFromName)(keyTypeName);
                if (keyType === undefined) {
                    const msg = `a Map type property "${propName}" of the type "${typeName}" has non-string key type: "${keyTypeName}"`;
                    if (this.shouldLenientTypeCheck) {
                        (0, logger_1.warn)(msg);
                        return undefined;
                    }
                    throw new Error(msg);
                }
                const valueTypeName = kvType === null || kvType === void 0 ? void 0 : kvType[1];
                const valueType = (0, dynamodb_primitive_types_1.dynamodbPrimitiveTypeFromName)(valueTypeName);
                if (valueType === undefined) {
                    const msg = `a property "${propName}" of the type "${typeName}" has unsupported type: "${valueTypeName}"`;
                    if (this.shouldLenientTypeCheck) {
                        (0, logger_1.warn)(msg);
                        return undefined;
                    }
                    throw new Error(msg);
                }
                return new dynamodb_item_field_1.KeyValuePairMapField(propName, keyType, valueType, this.shouldLenientTypeCheck);
            }
            if (valueDeclSymbolName === 'Uint8Array') {
                return new dynamodb_item_field_1.PrimitiveField(propName, new dynamodb_primitive_types_1.DynamodbPrimitiveType(dynamodb_primitive_types_1.DynamodbPrimitiveTypeKinds.Binary, dynamodb_primitive_types_1.DynamodbPrimitiveTypeFlags.NA));
            }
            let colonTokenCame = false;
            for (const propNode of prop.valueDeclaration.getChildren()) {
                if (colonTokenCame) {
                    const fieldType = (0, dynamodb_primitive_types_1.dynamodbPrimitiveTypeFromName)(propNode.getText());
                    if (fieldType === undefined) {
                        const msg = `a property "${propName}" of the type "${typeName}" has unsupported type: "${propNode.getText()}"`;
                        if (this.shouldLenientTypeCheck) {
                            (0, logger_1.warn)(msg);
                            return undefined;
                        }
                        throw new Error(msg);
                    }
                    return new dynamodb_item_field_1.PrimitiveField(propName, fieldType);
                }
                colonTokenCame = propNode.kind == typescript_1.default.SyntaxKind.ColonToken;
            }
            const msg = `unexpected error: a property "${propName}" of the type "${typeName}" has unsupported type`;
            if (this.shouldLenientTypeCheck) {
                (0, logger_1.warn)(msg);
                return undefined;
            }
            throw new Error(msg);
        });
    }
    extractKeyValueTypesFromKeyValuePairMapSyntax(nodes) {
        for (const node of nodes) {
            if (node.kind === typescript_1.default.SyntaxKind.TypeLiteral && node.getChildCount() === 3) {
                const kvTypeDeclNode = node.getChildAt(1);
                if (kvTypeDeclNode.kind === typescript_1.default.SyntaxKind.SyntaxList && kvTypeDeclNode.getChildCount() === 1) {
                    const kvTypeSignatureNode = kvTypeDeclNode.getChildAt(0);
                    if (kvTypeSignatureNode.kind === typescript_1.default.SyntaxKind.IndexSignature && kvTypeSignatureNode.getChildCount() === 5) {
                        const valueType = kvTypeSignatureNode.getChildAt(4).getText();
                        const keyTypeDeclNode = kvTypeSignatureNode.getChildAt(1);
                        if (keyTypeDeclNode.kind === typescript_1.default.SyntaxKind.SyntaxList && keyTypeDeclNode.getChildCount() === 1) {
                            const keyTypeSignatureNode = keyTypeDeclNode.getChildAt(0);
                            if (keyTypeSignatureNode.kind === typescript_1.default.SyntaxKind.Parameter && keyTypeSignatureNode.getChildCount() === 3) {
                                return [keyTypeSignatureNode.getChildAt(2).getText(), valueType];
                            }
                        }
                    }
                }
                break;
            }
        }
        return undefined;
    }
    isPropertyModifierInArgumentSuitableForDynamodbAttr(modifiers) {
        if (modifiers === undefined) {
            return true;
        }
        return !modifiers.map(m => m.kind).includes(typescript_1.default.SyntaxKind.PrivateKeyword);
    }
}
exports.FieldsCollector = FieldsCollector;
